from copy import deepcopy
from datetime import datetime, timedelta
from unittest.mock import Mock, PropertyMock
from lowball.models.authentication_models.token import Token
import flask
import jwt
import pytest
from lowball.authentication import Authentication
from lowball.builtins import DefaultAuthDB, DefaultAuthProvider
from lowball.core import Lowball
from lowball.exceptions import InvalidCredentialsException, NotImplementedException
from lowball.models.provider_models.auth_provider import AuthPackage, AuthProvider, CreateClientPackage, \
    ClientRegistrationPackage, SelfUpdateClientPackage, UpdateClientPackage
from lowball.models.provider_models.auth_db import AuthDatabase
from lowball.models.config_models import Config
from lowball.models.authentication_models import ClientData
import json


###################
# Object Fixtures #
###################


####################
# Request Fixtures #
####################
@pytest.fixture(params=["", "{almost,json}", 1, {"client_id": "something", "BLOB": "something"}])
def does_not_match_auth_package_request(request):
    return request.param

@pytest.fixture
def does_not_match_self_register_package_request(does_not_match_auth_package_request):
    return does_not_match_auth_package_request

@pytest.fixture
def does_not_match_self_update_package_request(does_not_match_auth_package_request):
    return does_not_match_auth_package_request

@pytest.fixture
def does_not_match_update_package_request(does_not_match_auth_package_request):
    return does_not_match_auth_package_request

@pytest.fixture()
def success_request(username):
    return {
        "client_id": username,
        "client_secret": "password"
    }

@pytest.fixture
def basic_header(jwt_token):
    return {"Authorization": f"Bearer {jwt_token}"}

@pytest.fixture
def admin1_token1_request_header(token1_admin1, token_secret):
    return {"Authorization": f"Bearer {jwt.encode(token1_admin1.to_dict(), token_secret)}"}

@pytest.fixture
def normal1_token1_request_header(token1_normal1, token_secret):
    return {"Authorization": f"Bearer {jwt.encode(token1_normal1.to_dict(), token_secret)}"}

@pytest.fixture
def normal2_token1_request_header(token1_normal2, token_secret):
    return {"Authorization": f"Bearer {jwt.encode(token1_normal2.to_dict(), token_secret)}"}


@pytest.fixture(params=[0,1,2,3])
def not_authenticated_or_token_expired_header(request, token3_admin1_expired, token3_normal1_expired, token_secret):
    round = request.param
    if round == 0:
        return {}
    elif round == 1:
        return {"Authorization": f"Bearer InvalidToken"}
    elif round == 2:
        return {"Authorization": f"Bearer {jwt.encode(token3_admin1_expired.to_dict(), token_secret)}"}
    elif round == 3:
        return {"Authorization": f"Bearer {jwt.encode(token3_normal1_expired.to_dict(), token_secret)}"}

@pytest.fixture(params=[0,1,2,3])
def non_admin_headers(request, token1_normal1, token2_normal1, token1_normal2, token2_normal2, token_secret):
    fixture_round = request.param
    if fixture_round == 0:
        return {"Authorization": f"Bearer {jwt.encode(token2_normal1.to_dict(), token_secret)}"}
    if fixture_round == 1:
        return {"Authorization": f"Bearer {jwt.encode(token1_normal2.to_dict(), token_secret)}"}
    if fixture_round == 2:
        return {"Authorization": f"Bearer {jwt.encode(token2_normal2.to_dict(), token_secret)}"}
    if fixture_round == 3:
        return {"Authorization": f"Bearer {jwt.encode(token1_normal1.to_dict(), token_secret)}"}

@pytest.fixture(params=[0,1,2,3])
def token_request_rounds(token_secret, token1_normal1, token2_normal1, token1_admin1, token1_admin2, request):
    fixture_round = request.param

    if fixture_round == 0:
        return {"Authorization": f"Bearer {jwt.encode(token1_normal1.to_dict(), token_secret)}"}, token1_normal1
    if fixture_round == 1:
        return {"Authorization": f"Bearer {jwt.encode(token2_normal1.to_dict(), token_secret)}"}, token2_normal1
    if fixture_round == 2:
        return {"Authorization": f"Bearer {jwt.encode(token1_admin1.to_dict(), token_secret)}"}, token1_admin1
    if fixture_round == 3:
        return {"Authorization": f"Bearer {jwt.encode(token1_admin2.to_dict(), token_secret)}"}, token1_admin2


@pytest.fixture(params=[0, 1, 2, 3, 4, 5])
def invalid_update_roles_requests_rounds(request):

    request_round = request.param

    if request_round == 0:
        return {}
    if request_round == 1:
        return "string not json"
    if request_round == 2:
        return {"add": "not a list"}
    if request_round == 3:
        return {"add": [1, 2, 3]}
    if request_round == 4:
        return {"remove": "not a list"}
    if request_round == 5:
        return {"remove": [1, 2, 3]}


@pytest.fixture
def default_token_life():
    return 3600

@pytest.fixture
def max_token_life():
    return 2592000


@pytest.fixture(params=[0, 1, 2])
def admin_client_create_token_success_rounds(request, client_id_normal1, client_id_normal2, token_secret, fake_now,
                                             base_role1, base_role2, admin_role, all_roles, monkeypatch,
                                             token_id, client_id_admin1, max_token_life, default_token_life):

    request_round = request.param

    if request_round == 0:
        body = {
            "client_id": client_id_normal1,
            "roles": all_roles
        }
        test_token = Token(
            cid=client_id_normal1,
            r=all_roles,
            cts=fake_now,
            ets=fake_now + timedelta(seconds=default_token_life),
            rcid=client_id_admin1,
            tid=token_id)
        token_string = jwt.encode(test_token.to_dict(), token_secret)

        monkeypatch.setattr(
            Authentication,
            "create_token",
            Mock(wraps=Authentication.create_token,
                 return_value=(token_string, test_token)
                 ))

        return body, token_string, test_token

    if request_round == 1:
        body = {
            "client_id": client_id_normal2,
            "roles": base_role2,
            "token_life": 120
        }
        test_token = Token(
            cid=client_id_normal2,
            r=base_role2,
            cts=fake_now,
            ets=fake_now + timedelta(seconds=120),
            rcid=client_id_admin1,
            tid=token_id)
        token_string = jwt.encode(test_token.to_dict(), token_secret)

        monkeypatch.setattr(
            Authentication,
            "create_token",
            Mock(wraps=Authentication.create_token,
                 return_value=(token_string, test_token)
                 ))

        return body, token_string, test_token

    if request_round == 2:
        body = {

        }
        test_token = Token(
            cid=client_id_admin1,
            r=[],
            cts=fake_now,
            ets=fake_now + timedelta(seconds=default_token_life),
            rcid=client_id_admin1,
            tid=token_id)
        token_string = jwt.encode(test_token.to_dict(), token_secret)

        monkeypatch.setattr(
            Authentication,
            "create_token",
            Mock(wraps=Authentication.create_token,
                 return_value=(token_string, test_token)
                 ))

        return body, token_string, test_token


@pytest.fixture(params=[0, 1, 2])
def normal_client_create_token_success_rounds(request, client_id_normal1, token_secret, base_role1, fake_now,
                                              monkeypatch, token_id, default_token_life):
    request_round = request.param

    if request_round == 0:
        body = {
            "roles": base_role1
        }
        test_token = Token(
            cid=client_id_normal1,
            r=base_role1,
            cts=fake_now,
            ets=fake_now + timedelta(seconds=default_token_life),
            rcid=client_id_normal1,
            tid=token_id)
        token_string = jwt.encode(test_token.to_dict(), token_secret)

        monkeypatch.setattr(
            Authentication,
            "create_token",
            Mock(wraps=Authentication.create_token,
                 return_value=(token_string, test_token)
                 ))

        return body, token_string, test_token

    if request_round == 1:
        body = {
            "token_life": 120
        }
        test_token = Token(
            cid=client_id_normal1,
            r=[],
            cts=fake_now,
            ets=fake_now + timedelta(seconds=120),
            rcid=client_id_normal1,
            tid=token_id)
        token_string = jwt.encode(test_token.to_dict(), token_secret)

        monkeypatch.setattr(
            Authentication,
            "create_token",
            Mock(wraps=Authentication.create_token,
                 return_value=(token_string, test_token)
                 ))

        return body, token_string, test_token

    if request_round == 2:
        body = {

        }
        test_token = Token(
            cid=client_id_normal1,
            r=[],
            cts=fake_now,
            ets=fake_now + timedelta(seconds=default_token_life),
            rcid=client_id_normal1,
            tid=token_id)
        token_string = jwt.encode(test_token.to_dict(), token_secret)

        monkeypatch.setattr(
            Authentication,
            "create_token",
            Mock(wraps=Authentication.create_token,
                 return_value=(token_string, test_token)
                 ))

        return body, token_string, test_token

@pytest.fixture
def basic_client_data(all_roles, username):
    return ClientData(username, all_roles)


@pytest.fixture(params=[
    0,
    1,
    2,
    3,
    4
])
def admin_list_token_roles_request_response(request, token_db, base_role1, base_role2, admin_role):
    role1 = base_role1[0]
    role2 = base_role2[0]
    adminrole = admin_role[0]
    queries = [
        f"?roles={role1}",
        f"?roles={role2}",
        f"?roles={adminrole}",
        f"?roles={role1}&roles={role2}",
        f"?roles={adminrole}&roles={role2}"
    ]

    request_round = request.param
    if request_round == 0:
        query = queries[request_round]
        expected_response = [token.to_dict() for token in token_db.values() if role1 in token.roles]
        return query, expected_response
    if request_round == 1:
        query = queries[request_round]
        expected_response = [token.to_dict() for token in token_db.values() if role2 in token.roles]
        return query, expected_response

    if request_round == 2:
        query = queries[request_round]
        expected_response = [token.to_dict() for token in token_db.values() if adminrole in token.roles]
        return query, expected_response

    if request_round == 3:
        query = queries[request_round]
        expected_response = [token.to_dict() for token in token_db.values() if role1 in token.roles or role2 in token.roles]
        return query, expected_response

    if request_round == 4:
        query = queries[request_round]
        expected_response = [token.to_dict() for token in token_db.values() if adminrole in token.roles or role2 in token.roles]
        return query, expected_response

@pytest.fixture(params=[
    0,
    1,
    2,
    3
])
def admin_list_token_client_request_response(request, token_db, client_id_normal1, client_id_admin1, client_id_normal2):
    queries = [
        f"?client_ids={client_id_normal1}",
        f"?client_ids={client_id_normal2}",
        f"?client_ids={client_id_normal1}&client_ids={client_id_admin1}",
        f"?client_ids={client_id_normal2}&client_ids={client_id_normal1}&client_ids={client_id_admin1}"
    ]
    request_round = request.param
    if request_round == 0:
        query = queries[request_round]
        expected_response = [token.to_dict() for token in token_db.values() if
                             token.client_id == client_id_normal1]
        return query, expected_response
    if request_round == 1:
        query = queries[request_round]
        expected_response = [token.to_dict() for token in token_db.values() if
                             token.client_id == client_id_normal2]
        return query, expected_response

    if request_round == 2:
        query = queries[request_round]
        expected_response = [token.to_dict() for token in token_db.values() if
                             token.client_id == client_id_normal1 or token.client_id == client_id_admin1]
        return query, expected_response

    if request_round == 3:
        query = queries[request_round]
        expected_response = [token.to_dict() for token in token_db.values() if
                             token.client_id == client_id_normal1 or token.client_id == client_id_admin1 or token.client_id == client_id_normal2]
        return query, expected_response


@pytest.fixture
def token_db_admin_exclude_expired_response(token_db, fake_now):
    return [token.to_dict() for token in token_db.values() if token.expiration > fake_now]


@pytest.fixture(params=[
    0,
    1,
    2,
])
def admin_list_token_multi_query_request_response(request, token_db, client_id_normal1, base_role1, fake_now):
    role1 = base_role1[0]

    queries = [
        f"?client_ids={client_id_normal1}&roles={role1}",
        f"?client_ids={client_id_normal1}&exclude_expired=yes",
        f"?client_ids={client_id_normal1}&roles={role1}&exclude_expired=yes",
    ]

    request_round = request.param
    if request_round == 0:
        query = queries[request_round]
        expected_response = [token.to_dict() for token in token_db.values() if
                             token.client_id == client_id_normal1 and role1 in token.roles]
        return query, expected_response
    if request_round == 1:
        query = queries[request_round]
        expected_response = [token.to_dict() for token in token_db.values() if
                             token.client_id == client_id_normal1 and token.expiration > fake_now]
        return query, expected_response

    if request_round == 2:
        query = queries[request_round]
        expected_response = [token.to_dict() for token in token_db.values() if
                             token.client_id == client_id_normal1 and token.expiration > fake_now and role1 in token.roles]
        return query, expected_response


@pytest.fixture
def non_admin_client_list_token_expected_response(client_id_normal1, token_db):

    return [token.to_dict() for token in token_db.values() if token.client_id == client_id_normal1]

@pytest.fixture(params=[
    0,
    1,
    2,
    3,
    4
])
def admin_delete_token_roles_request_response(request, token_db, base_role1, base_role2, admin_role):
    role1 = base_role1[0]
    role2 = base_role2[0]
    adminrole = admin_role[0]
    queries = [
        f"?roles={role1}",
        f"?roles={role2}",
        f"?roles={adminrole}",
        f"?roles={role1}&roles={role2}",
        f"?roles={adminrole}&roles={role2}"
    ]

    request_round = request.param
    if request_round == 0:
        query = queries[request_round]
        expected_response = [token for token in token_db.values() if role1 in token.roles]
        return query, expected_response
    if request_round == 1:
        query = queries[request_round]
        expected_response = [token for token in token_db.values() if role2 in token.roles]
        return query, expected_response

    if request_round == 2:
        query = queries[request_round]
        expected_response = [token for token in token_db.values() if adminrole in token.roles]
        return query, expected_response

    if request_round == 3:
        query = queries[request_round]
        expected_response = [token for token in token_db.values() if role1 in token.roles or role2 in token.roles]
        return query, expected_response

    if request_round == 4:
        query = queries[request_round]
        expected_response = [token for token in token_db.values() if adminrole in token.roles or role2 in token.roles]
        return query, expected_response


@pytest.fixture(params=[
    0,
    1,
    2,
    3
])
def admin_delete_token_client_request_response(request, token_db, client_id_normal1, client_id_admin1,
                                               client_id_normal2):
    queries = [
        f"?client_ids={client_id_normal1}",
        f"?client_ids={client_id_normal2}",
        f"?client_ids={client_id_normal1}&client_ids={client_id_admin1}",
        f"?client_ids={client_id_normal2}&client_ids={client_id_normal1}&client_ids={client_id_admin1}"
    ]
    request_round = request.param
    if request_round == 0:
        query = queries[request_round]
        expected_response = [token for token in token_db.values() if
                             token.client_id == client_id_normal1]
        return query, expected_response
    if request_round == 1:
        query = queries[request_round]
        expected_response = [token for token in token_db.values() if
                             token.client_id == client_id_normal2]
        return query, expected_response

    if request_round == 2:
        query = queries[request_round]
        expected_response = [token for token in token_db.values() if
                             token.client_id == client_id_normal1 or token.client_id == client_id_admin1]
        return query, expected_response

    if request_round == 3:
        query = queries[request_round]
        expected_response = [token for token in token_db.values() if
                             token.client_id == client_id_normal1 or token.client_id == client_id_admin1 or token.client_id == client_id_normal2]
        return query, expected_response


@pytest.fixture()
def admin_delete_token_multi_query_request_response(token_db, client_id_normal1, base_role1, fake_now):
    role1 = base_role1[0]

    query = f"?client_ids={client_id_normal1}&roles={role1}"

    expected_response = [token for token in token_db.values() if
                         token.client_id == client_id_normal1 and role1 in token.roles]
    return query, expected_response


@pytest.fixture(params=[
    0,
    1,
    2
])
def non_admin_client_list_token_expected_response_with_queries(client_id_normal1, fake_now,
                                                                  base_role1, base_role2, request, token_db):
    role1 = base_role1[0]
    role2 = base_role2[0]

    queries = [
        f"?roles={role1}",
        f"?roles={role1}&exclude_expired=yes",
        f"?roles={role2}"
    ]
    request_round = request.param
    if request_round == 0:
        query = queries[request_round]
        expected_response = [token.to_dict() for token in token_db.values() if
                             token.client_id == client_id_normal1 and role1 in token.roles]
        return query, expected_response
    if request_round == 1:
        query = queries[request_round]
        expected_response = [token.to_dict() for token in token_db.values() if
                             token.client_id == client_id_normal1 and token.expiration > fake_now]
        return query, expected_response

    if request_round == 2:
        query = queries[request_round]
        expected_response = [token.to_dict() for token in token_db.values() if
                             token.client_id == client_id_normal1 and role2 in token.roles]
        return query, expected_response

@pytest.fixture(params=[
    "tooshort",
    "toolongtobeatokenid",
    "goodlengthbut$$!",
    "NUMBER4TIME!RUN!"
])
def invalid_token_ids(request):
    return request.param


@pytest.fixture(params=[
    0,
    1,
    2
])
def expected_lookup_token_response_for_admin(request, token1_admin1, token2_admin2, token3_normal1_expired):
    request_round = request.param
    if request_round == 0:
        return token1_admin1.token_id, token1_admin1.to_dict()
    if request_round == 1:
        return token2_admin2.token_id, token2_admin2.to_dict()
    if request_round == 2:
        return token3_normal1_expired.token_id, token3_normal1_expired.to_dict()


@pytest.fixture(params=[
    0,
    1,
    2
])
def normal1_token_ids_request_response(request, token1_normal1, token2_normal1, token3_normal1_expired):

    request_round = request.param
    if request_round == 0:
        return token1_normal1.token_id, token1_normal1.to_dict()
    if request_round == 1:
        return token2_normal1.token_id, token2_normal1.to_dict()
    if request_round == 2:
        return token3_normal1_expired.token_id, token3_normal1_expired.to_dict()

@pytest.fixture(params=[0,1])
def get_client_request_response_rounds(request, admin1_token1_request_header,
                                       normal1_token1_request_header, client_data_normal1, client_data_admin1,
                                       client_id_admin1, client_id_normal1
                                       ):

    request_round = request.param

    if request_round == 0:
        return admin1_token1_request_header, client_id_admin1, client_data_admin1.to_dict()
    if request_round == 1:
        return admin1_token1_request_header, client_id_normal1, client_data_normal1.to_dict()

@pytest.fixture(params=[0,1])
def get_current_client_request_response_rounds(request, admin1_token1_request_header,
                                       normal1_token1_request_header, client_data_normal1, client_data_admin1,
                                       client_id_admin1, client_id_normal1
                                       ):

    request_round = request.param

    if request_round == 0:
        return admin1_token1_request_header, client_id_admin1, client_data_admin1.to_dict()
    if request_round == 1:
        return normal1_token1_request_header, client_id_normal1, client_data_normal1.to_dict()

@pytest.fixture(params=[0, 1, 2])
def get_clients_request_response_rounds(request, client_data_admin1, client_data_normal1, admin_role, base_role1):
    request_round = request.param
    admin_role_value = admin_role[0]
    role1 = base_role1[0]
    if request_round == 0:
        query = f"?roles={admin_role_value}"
        expected_client_data = [client_data_admin1]
        return query, expected_client_data
    if request_round == 1:
        query = f"?roles={role1}"
        expected_client_data = [client_data_normal1]
        return query, expected_client_data

    if request_round == 2:
        query = f"?roles={role1}&roles={admin_role_value}"
        expected_client_data = [client_data_admin1, client_data_normal1]
        return query, expected_client_data




@pytest.fixture(params=[0, 1, 2, 3])
def create_token_request_bad_token_life_rounds(request):
    request_round = request.param
    if request_round == 0:
        return {
            "token_life": "not an integer",
            "client_id": "test",
            "roles": []
        }
    if request_round == 1:
        return {
            "token_life": 8,
            "client_id": "test",
            "roles": []
        }
    if request_round == 2:
        return {
            "token_life": 999999999999999,
            "client_id": "test",
            "roles": []
        }
    if request_round == 3:
        return {
            "token_life": True,
            "client_id": "test",
            "roles": []
        }




###################
# Client Fixtures #
###################

@pytest.fixture
def test_app_config(token_secret, default_token_life, max_token_life):

    config = Config()
    config.authentication.token_secret = token_secret
    config.authentication.default_token_life = default_token_life
    config.authentication.max_token_life = max_token_life
    return config


@pytest.fixture
def lowball_app(test_app_config):
    app = Lowball(response_class=flask.Response, logging_handler=None, config=test_app_config)
    app.config["TESTING"] = True
    app.logger._log = Mock()

    with app.test_client() as client:
        yield client


@pytest.fixture
def lowball_app_mock_providers(test_authdb_base, test_authprovider_base, test_app_config):
    app = Lowball(response_class=flask.Response, logging_handler=None,
                  auth_provider=test_authprovider_base, auth_database=test_authdb_base,
                  config=test_app_config)
    app.config["TESTING"] = True
    app.logger._log = Mock()

    with app.test_client() as client:
        yield client


@pytest.fixture
def lowball_app_no_providers(test_app_config):
    app = Lowball(response_class=flask.Response,
                  logging_handler=None, auth_provider=None,
                  auth_database=None, config=test_app_config)
    app.config["TESTING"] = True
    app.logger._log = Mock()

    with app.test_client() as client:
        yield client


@pytest.fixture
def lowball_app_no_db(test_authprovider_base, test_app_config):
    app = Lowball(response_class=flask.Response, logging_handler=None, auth_database=None,
                  auth_provider=test_authprovider_base, config=test_app_config)
    app.config["TESTING"] = True
    app.logger._log = Mock()

    with app.test_client() as client:
        yield client


@pytest.fixture
def lowball_app_no_auth_provider(test_authdb_base, test_app_config):
    app = Lowball(response_class=flask.Response, logging_handler=None, auth_provider=None,
                  auth_database=test_authdb_base, config=test_app_config)
    app.config["TESTING"] = True
    app.logger._log = Mock()

    with app.test_client() as client:
        yield client

##############################
# Expected Response Fixtures #
##############################

@pytest.fixture
def expected_login_error_response():
    return "400 Bad Request: The Authentication Request Did Not Supply The Required Data"

@pytest.fixture
def expected_bad_package_data_error_response():
    return "400 Bad Request: The Request Did Not Supply The Required Data for the Auth Provider"

@pytest.fixture
def expected_no_auth_provider_error_response():
    return "503 Service Unavailable: No Authentication Provider Present"

@pytest.fixture
def expected_inadequate_roles_exception_response():
    return "401 Unauthorized: Current Token Has Inadequate Roles for Requested Action"


@pytest.fixture
def expected_non_list_returned_from_list_tokens_response():
    return "500 Internal Server Error: auth_db.list_tokens did not return a list as expected"


@pytest.fixture
def expected_non_token_in_token_list_response():
    return "500 Internal Server Error: auth_db.list_tokens returned a list that included a non-Token object"


@pytest.fixture
def expected_nonexistent_token_response():
    return "404 Not Found: Specified token not found"


@pytest.fixture
def expected_no_auth_db_response():
    return "503 Service Unavailable: No authentication database configured for this application"


@pytest.fixture
def expected_bad_token_life_response():
    return "token_life must be a positive integer greater than 10 and less than 2592000"

###############
# Basic Mocks #
###############

@pytest.fixture(params=[
    TypeError("TESTEXCEPTION"),
    ValueError("TESTEXCEPTION"),
    Exception("TESTEXCEPTION")
])
def mock_initialized_exception(monkeypatch, request):
    monkeypatch.setattr(TestAuthProvider, "initialized", Mock(side_effect=request.param))


@pytest.fixture(params=[1, 1.1, object(), [], {}, "string"])
def mock_initialized_non_bool(monkeypatch, request):
    monkeypatch.setattr(TestAuthProvider, "initialized", request.param)


@pytest.fixture(params=[True, False])
def mock_initialized_bool(monkeypatch, request):
    monkeypatch.setattr(TestAuthProvider, "initialized", request.param)
    return request.param

########################
# Authentication Mocks #
########################
@pytest.fixture
def mock_create_token(monkeypatch, token, jwt_token):

    monkeypatch.setattr(Authentication, "create_token", Mock(return_value=(jwt_token, token)))

@pytest.fixture
def mock_decode_token(monkeypatch, token):
    monkeypatch.setattr(Authentication, "decode_token", Mock(return_value=token))

################
# AuthDB Mocks #
################

@pytest.fixture
def token_db(token1_normal1, token2_normal1, token2_normal2, token1_normal2,
                             token1_admin2, token1_admin1, token2_admin1, token2_admin2, token,
                             token3_admin1_expired, token3_normal1_expired, token3_normal2_expired,
                             token3_admin2_expired):
    return {
            token.token_id: token,
            token1_normal1.token_id: token1_normal1,
            token2_normal1.token_id: token2_normal1,
            token2_normal2.token_id: token2_normal2,
            token1_normal2.token_id: token1_normal2,
            token1_admin2.token_id: token1_admin2,
            token1_admin1.token_id: token1_admin1,
            token2_admin1.token_id: token2_admin1,
            token2_admin2.token_id: token2_admin2,
            token3_admin1_expired.token_id: token3_admin1_expired,
            token3_normal1_expired.token_id: token3_normal1_expired,
            token3_normal2_expired.token_id: token3_normal2_expired,
            token3_admin2_expired.token_id: token3_admin2_expired,
        }

class TestAuthDB(AuthDatabase):



    def add_token(self, token_object):
        pass

    def lookup_token(self, token_id):

        pass

    def list_tokens(self):

        pass

    def list_tokens_by_role(self, role):
        pass

    def list_tokens_by_client_id(self, client_id):

        pass

    def cleanup_tokens(self):

        pass

    def revoke_all(self):
        pass

    def revoke_token(self, token_id):
        pass

@pytest.fixture
def test_authdb_base():

    return TestAuthDB

@pytest.fixture
def mock_lookup_token(monkeypatch, token):
    monkeypatch.setattr(TestAuthDB, "lookup_token", Mock(return_value=token))

@pytest.fixture
def mock_revoke_token(monkeypatch, token):
    monkeypatch.setattr(TestAuthDB, "revoke_token", Mock())

@pytest.fixture
def mock_lookup_token_filled(monkeypatch, token_db):
    def lookup_token(token_id):

        return token_db.get(token_id)

    monkeypatch.setattr(TestAuthDB, "lookup_token", Mock(wraps=lookup_token))

@pytest.fixture
def mock_list_tokens_filled(monkeypatch, token_db):
    def list_tokens():
        return list(token_db.values())

    monkeypatch.setattr(TestAuthDB, "list_tokens", Mock(wraps=list_tokens))

@pytest.fixture
def mock_list_tokens_by_role(monkeypatch, token_db):
    def list_tokens_by_role(role):

        return [token for token in list(token_db.values()) if role in token.roles]

    monkeypatch.setattr(TestAuthDB, "list_tokens_by_role", Mock(wraps=list_tokens_by_role))

@pytest.fixture
def mock_list_tokens_by_client_id(monkeypatch, token_db):

    def list_tokens_by_client_id(client_id):
        return [token for token in list(token_db.values()) if client_id == token.client_id]

    monkeypatch.setattr(TestAuthDB, "list_tokens_by_client_id", Mock(wraps=list_tokens_by_client_id))

@pytest.fixture(params=[0, 1])
def mock_list_tokens_returns_bad_values(monkeypatch, request):
    request_round = request.param
    if request_round == 0:

        monkeypatch.setattr(TestAuthDB, "list_tokens", Mock(return_value=["not", "tokens"]))
        monkeypatch.setattr(TestAuthDB, "list_tokens_by_client_id", Mock(return_value=["not", "tokens"]))
    if request_round == 1:
        monkeypatch.setattr(TestAuthDB, "list_tokens", Mock(return_value="not list"))
        monkeypatch.setattr(TestAuthDB, "list_tokens_by_client_id", Mock(return_value="not list"))

@pytest.fixture
def mock_revoke_token(monkeypatch):
    monkeypatch.setattr(TestAuthDB, "revoke_token", Mock())

@pytest.fixture
def mock_revoke_all(monkeypatch):

    monkeypatch.setattr(TestAuthDB, "revoke_all", Mock())

@pytest.fixture
def mock_cleanup_tokens(monkeypatch):
    monkeypatch.setattr(TestAuthDB, "cleanup_tokens", Mock())

@pytest.fixture
def mock_add_token(monkeypatch):
    monkeypatch.setattr(TestAuthDB, "add_token", Mock())


#######################
# Auth Provider Mocks #
#######################

class TestAuthProvider(AuthProvider):

    def authenticate(self, auth_package):

        pass

    @property
    def auth_package_class(self):
        return None

@pytest.fixture
def client_data_normal1(client_id_normal1, base_role1):
    return ClientData(client_id=client_id_normal1, roles=base_role1)

@pytest.fixture
def client_data_admin1(client_id_admin1, admin_role):
    return ClientData(client_id=client_id_admin1, roles=admin_role)

@pytest.fixture
def client_db(client_id_normal1, client_id_admin1, client_data_normal1, client_data_admin1):

    return {
        client_id_normal1: client_data_normal1,
        client_id_admin1: client_data_admin1
    }

@pytest.fixture
def mock_get_client_filled(monkeypatch, client_db):

    def get_client(client_id):

        return client_db.get(client_id)

    monkeypatch.setattr(TestAuthProvider, "get_client", Mock(wraps=get_client))

@pytest.fixture
def mock_get_client_not_implemented(monkeypatch):

    monkeypatch.setattr(TestAuthProvider, "get_client", Mock(side_effect=NotImplementedException("get_client")))

@pytest.fixture
def mock_list_clients_filled(monkeypatch, client_db):

    def list_clients():

        return list(client_db.values())

    monkeypatch.setattr(TestAuthProvider, "list_clients", Mock(wraps=list_clients))

@pytest.fixture(params=[0, 1])
def mock_list_clients_returns_bad_values(monkeypatch, request):
    request_round = request.param
    if request_round == 0:

        monkeypatch.setattr(TestAuthProvider, "list_clients", Mock(return_value=["not", "client_data"]))
    if request_round == 1:
        monkeypatch.setattr(TestAuthProvider, "list_clients", Mock(return_value="not list"))

@pytest.fixture
def mock_delete_roles(monkeypatch):

    monkeypatch.setattr(TestAuthProvider, "delete_roles", Mock())

@pytest.fixture
def mock_add_roles(monkeypatch):

    monkeypatch.setattr(TestAuthProvider, "add_roles", Mock())

@pytest.fixture
def mock_get_client_not_found(monkeypatch):
    monkeypatch.setattr(TestAuthProvider, "get_client", Mock(return_value=None))

@pytest.fixture
def mock_get_client_found(monkeypatch, basic_client_data):
    monkeypatch.setattr(TestAuthProvider, "get_client", Mock(return_value=basic_client_data))

@pytest.fixture
def test_authprovider_base():

    return TestAuthProvider

@pytest.fixture
def mock_auth_package_class(monkeypatch):

    class TestAuthPackage(AuthPackage):
        def __init__(self, client_id, client_secret):
            self.client_id = client_id
            self.client_secret = client_secret

    monkeypatch.setattr(TestAuthProvider, "auth_package_class", PropertyMock(return_value=TestAuthPackage))

@pytest.fixture
def mock_self_register_class(monkeypatch):
    class TestSelfRegisterPackage(ClientRegistrationPackage):
        def __init__(self, client_id, client_secret):
            self.client_id = client_id
            self.client_secret = client_secret
    monkeypatch.setattr(TestAuthProvider, "client_registration_package_class", PropertyMock(return_value=TestSelfRegisterPackage))

@pytest.fixture
def mock_self_update_class(monkeypatch):
    class TestSelfUpdateClientPackage(SelfUpdateClientPackage):
        def __init__(self, client_id, client_secret):
            self.client_id = client_id
            self.client_secret = client_secret
    monkeypatch.setattr(TestAuthProvider, "self_update_client_package_class", PropertyMock(return_value=TestSelfUpdateClientPackage))


@pytest.fixture
def mock_update_client_class(monkeypatch):
    class TestUpdateClientPackage(UpdateClientPackage):
        def __init__(self, client_id, client_secret):

            self.client_id = client_id
            self.client_secret = client_secret

    monkeypatch.setattr(TestAuthProvider, "update_client_package_class",
                            PropertyMock(return_value=TestUpdateClientPackage))

@pytest.fixture
def mock_create_client_class(monkeypatch):
    class TestCreateClientPackage(CreateClientPackage):
        def __init__(self, client_id, client_secret):

            self.client_id = client_id
            self.client_secret = client_secret

    monkeypatch.setattr(TestAuthProvider, "create_client_package_class",
                            PropertyMock(return_value=TestCreateClientPackage))


@pytest.fixture
def mock_delete_client_simple(monkeypatch):
    monkeypatch.setattr(TestAuthProvider, "delete_client", Mock())

@pytest.fixture
def mock_enable_client_simple(monkeypatch):
    monkeypatch.setattr(TestAuthProvider, "enable_client", Mock())

@pytest.fixture
def mock_disable_client_simple(monkeypatch):
    monkeypatch.setattr(TestAuthProvider, "disable_client", Mock())

@pytest.fixture
def mock_add_and_delete_roles_simple(monkeypatch):

    monkeypatch.setattr(TestAuthProvider, "add_roles", Mock())
    monkeypatch.setattr(TestAuthProvider, "delete_roles", Mock())


@pytest.fixture(params=[0, 1, 2])
def mock_self_register_rounds(request, monkeypatch, basic_client_data):
    if request.param == 0:
        monkeypatch.setattr(TestAuthProvider, "client_self_register", Mock(return_value=basic_client_data))
    elif request.param == 1:
        monkeypatch.setattr(TestAuthProvider, "client_self_register", Mock(return_value=basic_client_data.to_dict()))

    elif request.param == 2:
        monkeypatch.setattr(TestAuthProvider, "client_self_register", Mock(return_value=json.dumps(basic_client_data.to_dict())))

@pytest.fixture(params=[0, 1, 2])
def mock_self_update_rounds(request, monkeypatch, basic_client_data):
    if request.param == 0:
        monkeypatch.setattr(TestAuthProvider, "client_self_update", Mock(return_value=basic_client_data))
    elif request.param == 1:
        monkeypatch.setattr(TestAuthProvider, "client_self_update", Mock(return_value=basic_client_data.to_dict()))

    elif request.param == 2:
        monkeypatch.setattr(TestAuthProvider, "client_self_update", Mock(return_value=json.dumps(basic_client_data.to_dict())))

@pytest.fixture(params=[0, 1, 2])
def mock_update_client_rounds(request, monkeypatch, basic_client_data):
    if request.param == 0:
        monkeypatch.setattr(TestAuthProvider, "update_client", Mock(return_value=basic_client_data))
    elif request.param == 1:
        monkeypatch.setattr(TestAuthProvider, "update_client", Mock(return_value=basic_client_data.to_dict()))

    elif request.param == 2:
        monkeypatch.setattr(TestAuthProvider, "update_client", Mock(return_value=json.dumps(basic_client_data.to_dict())))

@pytest.fixture(params=[0, 1, 2])
def mock_create_client_rounds(request, monkeypatch, basic_client_data):
    if request.param == 0:
        monkeypatch.setattr(TestAuthProvider, "create_client", Mock(return_value=basic_client_data))
    elif request.param == 1:
        monkeypatch.setattr(TestAuthProvider, "create_client", Mock(return_value=basic_client_data.to_dict()))

    elif request.param == 2:
        monkeypatch.setattr(TestAuthProvider, "create_client", Mock(return_value=json.dumps(basic_client_data.to_dict())))



@pytest.fixture
def mock_authenticate_basic(monkeypatch, username, all_roles):

    monkeypatch.setattr(TestAuthProvider, "authenticate", Mock(return_value=ClientData(username, roles=all_roles)))









from setuptools import setup, find_packages

DESCRIPTION = "Lowball is designed to add simple endpoint level RBAC to your Flask based API services."
VERSION = "1.0.2"


def read_requirements():
    """
    Simple helper method to read in the requirements.txt file and parse it into a list for consumption by setuptools
    :return: list of requirements
    """
    required = []
    with open("requirements.txt") as f:
        for line in f:
            if line[0] != "#":
                # Not a comment add it as a requirement
                required.append(line.split("#"))
    return required


def readme():
    """
    Helper to try and format the .md readme file for pretty printing. Falls back to short description.
    :return: The available description
    """
    try:
        import pypandoc
        description = pypandoc.convert_file('README.md', 'rst')
    except:
        description = DESCRIPTION

    return description


setup(name="lowball",
      version=VERSION,
      description=DESCRIPTION,
      long_description=readme(),
      url="https://github.com/EmersonElectricCo/lowball",
      author="Isaiah Eichen, Grant Steiner, Timothy Lemm",
      author_email="compsecmonkey@gmail.com",
      license="Apache License 2.0",
      packages=find_packages(),
      install_requires=read_requirements(),
      test_suite="lowball.tests",
      zip_safe=False
      )



import pytest
import datetime
import uuid

from flask import Flask
from werkzeug.datastructures import Headers
import jwt

import lowball.models.config_models
from lowball.authentication.authentication import Authentication
from lowball.models.authentication_models import Token, ClientData, generate_token_id
from lowball.exceptions import InvalidTokenException
from unittest.mock import Mock
import lowball


@pytest.fixture
def bad_expiration():
    return datetime.datetime(2030, 1, 1)


@pytest.fixture(params=[1, "string", True, object(), {}, [], None])
def not_auth_config(request):
    return request.param


@pytest.fixture(params=[1, "string", True, object(), {}, []])
def not_logger(request):
    return request.param


@pytest.fixture(autouse=True)
def patched_uuid4(monkeypatch, token_id):
    monkeypatch.setattr(uuid, "uuid4", lambda: token_id)

@pytest.fixture(autouse=True)
def patched_generate_token(token_id, monkeypatch):
    monkeypatch.setattr(lowball.authentication.authentication, "generate_token_id", Mock(return_value=token_id))



@pytest.fixture
def create_token1(token_id, fake_now, all_roles):
    return Token(
        cid="username1",
        r=all_roles,
        cts=fake_now,
        ets=datetime.datetime(2020, 2, 1),
        rcid="username2",
        tid=token_id
    )

@pytest.fixture
def create_token2(token_id, fake_now, all_roles):
    return Token(
        cid="username1",
        r=all_roles,
        cts=fake_now,
        ets=datetime.datetime(2020, 2, 1),
        rcid="username1",
        tid=token_id
    )

@pytest.fixture
def create_token3(token_id, fake_now, all_roles):
    return Token(
        cid="username1",
        r=all_roles,
        cts=fake_now,
        ets=datetime.datetime(2020, 1, 1, 0, 30),
        rcid="username2",
        tid=token_id
    )

@pytest.fixture(params=["requesting_user", "no_requesting_user", "no_expiration"])
def token_create_rounds(request, client_data, create_token1, create_token2, create_token3, token_secret):
    # return
    if request.param == "requesting_user":

        return client_data, "username2", datetime.datetime(2020, 2, 1), datetime.datetime(2020, 2, 1), jwt.encode(create_token1.to_dict(), token_secret)

    if request.param == "no_requesting_user":

        return client_data, None, datetime.datetime(2020, 2, 1), datetime.datetime(2020, 2, 1), jwt.encode(create_token2.to_dict(), token_secret)

    if request.param == "no_expiration":

        return client_data, "username2", None, datetime.datetime(2020, 1, 1, 0, 30), jwt.encode(create_token3.to_dict(), token_secret)

@pytest.fixture(params=[0,1,2])
def bad_token_rounds(request, invalid_expected_token_string):
    if request.param == 0:
        return "notevenkindofatoken"
    elif request.param == 1:
        return invalid_expected_token_string
    elif request.param == 2:
        return ["not", "even", "a", "string"]

@pytest.fixture
def decoded_data(token_id):
    return Token(
        cid="username1",
        r=['admin', 'lead', 'analyst'],
        cts='2020-01-01 00:00:00',
        ets='2020-01-01 00:30:00',
        rcid="username1",
        tid=token_id
    )


@pytest.fixture(params=[None, "username1"])
def requesting_user(request):
    return request.param


@pytest.fixture(params=[datetime.datetime(2020, 1, 2), "2020-01-02 00:00:00"])
def expiration(request):
    return request.param


@pytest.fixture(params=[None, 1, True, "string", {}, [], object()])
def not_auth_data(request):
    return request.param


@pytest.fixture(params=[1, True, {}, [], object()])
def not_user(request):
    return request.param


@pytest.fixture
def request_no_headers(client, headers):
    with client.test_client():
        with client.test_request_context(headers=headers) as client:
            yield client


@pytest.fixture
def request_invalid_headers(client, headers_invalid_auth):
    with client.test_client():
        with client.test_request_context(headers=headers_invalid_auth) as client:
            yield client


@pytest.fixture
def request_token_not_in_db(client, headers_with_token, token_id, monkeypatch):
    def mocked_decode_token(self, token):
        return Token(
            cid="username",
            r=[],
            cts=datetime.datetime(2019, 11, 1),
            ets=datetime.datetime(2020, 12, 1),
            rcid="another_user",
            tid=token_id
        )

    def mocked_lookup_token(token):
        raise InvalidTokenException

    client.auth_db.lookup_token = mocked_lookup_token

    with client.test_client():
        with client.test_request_context(headers=headers_with_token) as client:
            monkeypatch.setattr(Authentication, "decode_token", mocked_decode_token)
            yield client


@pytest.fixture
def request_unmatched_token(client, headers_with_token, token_id, monkeypatch):
    def mocked_decode_token(self, token):
        return Token(
            cid="username",
            r=[],
            cts=datetime.datetime(2019, 11, 1),
            ets=datetime.datetime(2020, 12, 1),
            rcid="another_user",
            tid=token_id
        )

    def mocked_lookup_token(token):
        return Token(
            cid="different_user",
            r=[],
            cts=datetime.datetime(2019, 11, 1),
            ets=datetime.datetime(2020, 12, 1),
            rcid="another_user",
            tid=token_id
        )

    client.auth_db.lookup_token = mocked_lookup_token

    with client.test_client():
        with client.test_request_context(headers=headers_with_token) as client:
            monkeypatch.setattr(Authentication, "decode_token", mocked_decode_token)
            yield client


@pytest.fixture
def request_valid_token(client, headers_with_token, token_id, monkeypatch, expected_token):
    def mocked_decode_token(self, token):
        return expected_token

    def mocked_lookup_token(token):
        return expected_token

    client.auth_db.lookup_token = mocked_lookup_token

    with client.test_client():
        with client.test_request_context(headers=headers_with_token) as client:
            monkeypatch.setattr(Authentication, "decode_token", mocked_decode_token)
            yield client


@pytest.fixture
def request_token_expired(client, headers_with_token, token_id, monkeypatch):
    def mocked_decode_token(self, token):
        return Token(
            cid="username",
            r=[],
            cts=datetime.datetime(2019, 11, 1),
            ets=datetime.datetime(2019, 12, 1),
            rcid="another_user",
            tid=token_id
        )

    with client.test_client():
        with client.test_request_context(headers=headers_with_token) as client:
            monkeypatch.setattr(Authentication, "decode_token", mocked_decode_token)
            yield client


@pytest.fixture
def request_token_no_roles(client, headers_with_token, token_id, monkeypatch, expected_token):
    def mocked_decode_token(self, token):
        return expected_token

    def mocked_lookup_token(token):
        return expected_token

    client.auth_db.lookup_token = mocked_lookup_token

    with client.test_client():
        with client.test_request_context(headers=headers_with_token) as client:
            monkeypatch.setattr(Authentication, "decode_token", mocked_decode_token)
            yield client


@pytest.fixture(params=[
    ["admin"],
    ["admin", "lead"],
    ["admin", "lead", "analyst"],
    ["admin", "analyst"],
    ["lead"],
    ["lead", "analyst"],
    ["analyst"]
])
def request_admin_or_lead_or_analyst_token(request, client, headers_with_token, expected_token, monkeypatch):
    expected_token.roles = request.param

    def mocked_decode_token(self, token):
        return expected_token

    def mocked_lookup_token(token):
        return expected_token

    client.auth_db.lookup_token = mocked_lookup_token

    with client.test_client():
        with client.test_request_context(headers=headers_with_token) as client:
            monkeypatch.setattr(Authentication, "decode_token", mocked_decode_token)
            yield client


@pytest.fixture(params=[
    ["admin"],
    ["admin", "lead"],
    ["admin", "lead", "analyst"],
    ["admin", "analyst"]
])
def request_admin_token(request, client, headers_with_token, expected_token, monkeypatch):
    expected_token.roles = request.param
    
    def mocked_decode_token(self, token):
        return expected_token

    def mocked_lookup_token(token):
        return expected_token

    client.auth_db.lookup_token = mocked_lookup_token
    
    with client.test_client():
        with client.test_request_context(headers=headers_with_token) as client:
            monkeypatch.setattr(Authentication, "decode_token", mocked_decode_token)
            yield client


@pytest.fixture(params=[
    ["lead"],
    ["lead", "analyst"],
    ["analyst"]
])
def request_non_admin_token(request, client, headers_with_token, expected_token, monkeypatch):
    expected_token.roles = request.param

    def mocked_decode_token(self, token):
        return expected_token

    def mocked_lookup_token(token):
        return expected_token

    client.auth_db.lookup_token = mocked_lookup_token

    with client.test_client():
        with client.test_request_context(headers=headers_with_token) as client:
            monkeypatch.setattr(Authentication, "decode_token", mocked_decode_token)
            yield client


@pytest.fixture(params=[
    ["admin"],
    ["admin", "lead"],
    ["admin", "lead", "analyst"],
    ["admin", "analyst"],
    ["lead"],
    ["lead", "analyst"],
    ["analyst"],
    []
])
def request_any_role_token(request, client, headers_with_token, expected_token, monkeypatch):
    expected_token.roles = request.param

    def mocked_decode_token(self, token):
        return expected_token

    def mocked_lookup_token(token):
        return expected_token

    client.auth_db.lookup_token = mocked_lookup_token

    with client.test_client():
        with client.test_request_context(headers=headers_with_token) as client:
            monkeypatch.setattr(Authentication, "decode_token", mocked_decode_token)
            yield client


@pytest.fixture(params=[
    ["admin"],
    ["admin", "lead"],
    ["admin", "analyst"],
    ["lead"],
    ["lead", "analyst"],
    ["analyst"],
    []
])
def request_not_all_roles_token(request, client, headers_with_token, expected_token, monkeypatch):
    expected_token.roles = request.param

    def mocked_decode_token(self, token):
        return expected_token

    def mocked_lookup_token(token):
        return expected_token

    client.auth_db.lookup_token = mocked_lookup_token

    with client.test_client():
        with client.test_request_context(headers=headers_with_token) as client:
            monkeypatch.setattr(Authentication, "decode_token", mocked_decode_token)
            yield client


@pytest.fixture
def request_all_roles_token(client, headers_with_token, expected_token, monkeypatch):
    expected_token.roles = ["admin", "lead", "analyst"]

    def mocked_decode_token(self, token):
        return expected_token

    def mocked_lookup_token(token):
        return expected_token

    client.auth_db.lookup_token = mocked_lookup_token

    with client.test_client():
        with client.test_request_context(headers=headers_with_token) as client:
            monkeypatch.setattr(Authentication, "decode_token", mocked_decode_token)
            yield client


@pytest.fixture()
def client_without_authdb(auth):
    test_app = Flask("test")
    test_app.config["TESTING"] = True
    test_app.auth_db = None
    test_app.authenticator = auth
    return test_app


@pytest.fixture
def request_with_no_authdb(client_without_authdb, headers_with_token):
    with client_without_authdb.test_client():
        with client_without_authdb.test_request_context(headers=headers_with_token) as client:
            yield client



from flask import Response, jsonify


class LowballResponse(Response):
    """
    Child class of Flask Response object used to create Default JSON Responses
    """

    @classmethod
    def force_type(cls, return_value, environ=None):
        """Force all native Python types into proper responses.

        Extension of Parent class force_type to intercept and convert dict objects to json compliant objects

        :param return_value: data that is being returned from the view function.
        :type environ: dict
        :param environ: @super
        :rtype: LowballResponse
        :return: Response object
        """
        # These types are JSON-compliant, just pass them to jsonify
        if isinstance(return_value, (dict, list, tuple)):
            return jsonify(return_value)

        # These types need to be converted to lists before being jsonified
        elif isinstance(return_value, (set, frozenset, range)):
            return jsonify(list(return_value))
        
        # Convert these types to strings and set them as response data
        elif isinstance(return_value, (int, float, complex)):
            return LowballResponse(response=str(return_value))

        # Convert memoryview to bytes and set as response data
        elif isinstance(return_value, memoryview):
            return LowballResponse(response=bytes(return_value))

        # For all other return types (this will only ever be Response objects, or custom objects),
        # pass along data to base class `force_type`
        else:
            return super(LowballResponse, cls).force_type(response=return_value, environ=environ)



from flask import g, request


def request_finished_log(sender, response, **extra):
    """Subscriber for request_finished signal

    This subscriber will log pertinent information from the
    response of a view function, using the application logger.
    Exception and normal responses are handled differently.

    :type sender: flask.Flask
    :param sender: the application that is sending the signal
    :type response: flask.Response
    :param response: The response from the view function
    :type extra: dict
    :param extra: container for any extra keywords that get passed in
    :rtype: None
    :return: None
    """
    try:
        request_environment = request.environ
    except AttributeError:
        request_environment = None

    if hasattr(g, "client_data"):
        client_data = {
            "requesting_client": g.client_data.client_id,
            "client_token_id": g.client_data.token_id
        }
    else:
        client_data = {}

    if request_environment:
        log_dict = {
            "user_agent": request_environment.get("HTTP_USER_AGENT", "None"),
            "src_ip": request_environment.get("HTTP_X_FORWARDED_FOR", request_environment.get("REMOTE_ADDR")),
            "http_method": request.method,
            "url": request_environment["werkzeug.request"].full_path,
            "status_code": response.status_code,
            "client_data": client_data
        }

        if hasattr(g, "response_is_exception") and g.response_is_exception:
            message = {"result": response.status, "error_information": g.response_exception_log_data}
            sender.logger.error(message, extra=log_dict)
        else:
            message = {"result": response.status}
            sender.logger.info(message, extra=log_dict)


__all__ = [
    "request_finished_log"
]



from flask import g

from lowball.builtins.error_handler import default_error_handler
from lowball.builtins.response_class import LowballResponse


class TestDefaultErrorHandler:
    def test_basic_exception_handled_properly(self, basic_exception, client_with_response_class):
        response = default_error_handler(basic_exception)
        assert isinstance(response, LowballResponse)
        assert response.status_code == 500
        assert g.response_is_exception
        assert g.response_exception_log_data == {"error_type": str(type(basic_exception)),
                                        "error_msg": str(basic_exception)}

    def test_lowball_exception_is_handled_properly(self, lowball_exception, client_with_response_class):
        response = default_error_handler(lowball_exception)
        assert isinstance(response, LowballResponse)
        assert response.status_code == lowball_exception.code
        assert g.response_is_exception
        assert g.response_exception_log_data == lowball_exception.additional_log_data

    def test_http_exception_code_less_than_500_handled_properly(self, http_exception_lt_500,
                                                                client_with_response_class):
        response = default_error_handler(http_exception_lt_500)
        assert isinstance(response, LowballResponse)
        assert response.status_code == http_exception_lt_500.code

    def test_http_exception_greater_than_equal_500_handled_properly(self, http_exception_gte_500,
                                                                    client_with_response_class):
        response = default_error_handler(http_exception_gte_500)
        assert isinstance(response, LowballResponse)
        assert response.status_code == http_exception_gte_500.code
        assert g.response_is_exception
        assert g.response_exception_log_data == {"error_type": str(type(http_exception_gte_500)),
                                        "error_msg": str(http_exception_gte_500)}



from datetime import datetime

import pytest

from lowball.models.authentication_models import Token, valid_token_id, generate_token_id, TOKEN_ID_PATTERN
import re


class TestValidTokenID:

    @pytest.mark.parametrize("test_token_id,is_valid", [
        ("a"*16, True),
        ("a"*15, False),
        ("a"*17, False),
        ("0123456789abcdef", True),
        (None, False),
        (44747474744, False),
        ("aaaaaaaaaaaaaaa&", False)
    ])
    def test_valid_token_id_validates_tokens_correctly(self, test_token_id, is_valid):

        assert valid_token_id(test_token_id) == is_valid

class TestGenerateTokenID:

    def test_generate_token_id_creates_expected_string_format(self):
        for i in range(0, 1000):
            assert re.match(TOKEN_ID_PATTERN, generate_token_id()) is not None



class TestToken:
    @pytest.mark.parametrize("user, roles, created, expiration, issued_by, token_id", [
        (1, ["list", "of", "roles"], datetime(2020, 1, 1), datetime(2020, 2, 1), "issuer", "token_id"),
        ("user", 1, datetime(2020, 1, 1), datetime(2020, 2, 1), "issuer", "token_id"),
        ("user", ["list", "of", "roles"], 1, datetime(2020, 2, 1), "issuer", "token_id"),
        ("user", ["list", "of", "roles"], datetime(2020, 1, 1), 1, "issuer", "token_id"),
        ("user", ["list", "of", "roles"], datetime(2020, 1, 1), datetime(2020, 2, 1), 1, "token_id"),
        ("user", ["list", "of", "roles"], datetime(2020, 1, 1), datetime(2020, 2, 1), "issuer", 1),
    ])
    def test_init_raises_type_error_when_improper_data_passed_in(self, user, roles, created, expiration, issued_by,
                                                                 token_id):
        with pytest.raises(TypeError):
            Token(cid=user, r=roles, cts=created, ets=expiration, rcid=issued_by,
                  tid=token_id)

    def test_not_all_roles_strings_raises_value_error(self, username, created, expiration, issued_by, token_id, 
                                                      invalid_roles):
        with pytest.raises(ValueError):
            Token(cid=username, r=invalid_roles, cts=created, ets=expiration, rcid=issued_by,
                  tid=token_id)

    def test_to_dict_method_returns_proper_data(self, username, roles, created, expiration, issued_by, token_id,
                                                token_dict):
        token = Token(cid=username, r=roles, cts=created, ets=expiration, rcid=issued_by,
                      tid=token_id)
        assert token.to_dict() == token_dict

    def test_setting_user_to_bad_value_raises_type_error(self, username, roles, created, expiration, issued_by,
                                                         token_id, not_string):
        token = Token(cid=username, r=roles, cts=created, ets=expiration, rcid=issued_by,
                      tid=token_id)

        with pytest.raises(TypeError):
            token.client_id = not_string

    def test_setting_roles_to_bad_value_raises_type_error(self, username, roles, created, expiration, issued_by,
                                                          token_id, not_list):
        token = Token(cid=username, r=roles, cts=created, ets=expiration, rcid=issued_by,
                      tid=token_id)

        with pytest.raises(TypeError):
            token.roles = not_list
            
    def test_setting_roles_to_bad_value_after_init_raises_value_error(self, username, roles, created, expiration,
                                                                      issued_by, token_id, invalid_roles):
        token = Token(cid=username, r=roles, cts=created, ets=expiration, rcid=issued_by,
                      tid=token_id)

        with pytest.raises(ValueError):
            token.roles = invalid_roles

    def test_setting_created_to_not_datetime_raises_type_error(self, username, roles, created, expiration, issued_by,
                                                               token_id, not_datetime):
        token = Token(cid=username, r=roles, cts=created, ets=expiration, rcid=issued_by,
                      tid=token_id)

        with pytest.raises(TypeError):
            token.created = not_datetime

    def test_setting_expiration_to_not_datetime_raises_type_error(self, username, roles, created, expiration, issued_by,
                                                                  token_id, not_datetime):
        token = Token(cid=username, r=roles, cts=created, ets=expiration, rcid=issued_by,
                      tid=token_id)

        with pytest.raises(TypeError):
            token.expiration = not_datetime

    def test_setting_created_to_not_string_raises_type_error(self, username, roles, created, expiration, issued_by,
                                                             token_id, not_string):
        token = Token(cid=username, r=roles, cts=created, ets=expiration, rcid=issued_by,
                      tid=token_id)

        with pytest.raises(TypeError):
            token.created = not_string

    def test_setting_expiration_to_not_string_raises_type_error(self, username, roles, created, expiration, issued_by,
                                                                token_id, not_string):
        token = Token(cid=username, r=roles, cts=created, ets=expiration, rcid=issued_by,
                      tid=token_id)

        with pytest.raises(TypeError):
            token.expiration = not_string

    def test_setting_created_to_bad_string_raises_value_error(self, username, roles, created, expiration, issued_by,
                                                              token_id, bad_date_string):
        token = Token(cid=username, r=roles, cts=created, ets=expiration, rcid=issued_by,
                      tid=token_id)

        with pytest.raises(ValueError):
            token.created = bad_date_string

    def test_setting_expiration_to_bad_string_raises_value_error(self, username, roles, created, expiration, issued_by,
                                                                 token_id, bad_date_string):
        token = Token(cid=username, r=roles, cts=created, ets=expiration, rcid=issued_by,
                      tid=token_id)

        with pytest.raises(ValueError):
            token.expiration = bad_date_string

    def test_setting_issued_by_to_bad_value_raises_type_error(self, username, roles, created, expiration, issued_by,
                                                              token_id, not_string):
        token = Token(cid=username, r=roles, cts=created, ets=expiration, rcid=issued_by,
                      tid=token_id)

        with pytest.raises(TypeError):
            token.issued_by = not_string

    def test_setting_token_id_to_bad_value_raises_type_error(self, username, roles, created, expiration, issued_by,
                                                             token_id, not_string):
        token = Token(cid=username, r=roles, cts=created, ets=expiration, rcid=issued_by,
                      tid=token_id)

        with pytest.raises(TypeError):
            token.token_id = not_string

    @pytest.mark.parametrize("test_token_id", [
        ("a" * 15),
        ("a" * 17),
        ("aaaaaaaaaaaaaaa&")
    ])
    def test_setting_token_id_to_bad_string_raises_value_error(self, token, test_token_id):

        with pytest.raises(ValueError):
            token.token_id = test_token_id


from .logging.logging_handler import DefaultLoggingHandler
from .providers import DefaultAuthDB, DefaultAuthProvider
from .error_handler import default_error_handler
from .response_class import LowballResponse
from .response_class import LowballResponse
from .signal_subscribers import request_finished_log



"""
LOWBALL BUILTIN ROUTES

/builtins

- /auth - post/del/get (login/logout/current token info)

- /auth/tokens - get (current user all), post (create token), del (delete all current user tokens)

- /auth/tokens/all - get (all tokens in system) Admin, delete (revoke all tokens in system) (query params for client/roles/etc...)

- /auth/tokens/cleanup - post (admin) cleanup expired tokens

- /auth/tokens/<token id> - get/del ( admin can view/delete all, non admin only own)

- /auth/clients - get/post (get get self client, self update)

- /auth/clients/create - post (admin) create client

- /auth/clients/register - post client self registration

- /auth/clients/all get (admin) get all clients in system

- /auth/clients/<client_id> - (admin) get/delete/post (user info, delete user, update user)

- /auth/clients/<client_id>/roles delete (admin) /<role> - post/delete admin add/remove role from client

-/status

-/
"""

from .auth import auth
from .status import status